import re
import logging
import jsonpickle
from functools import partial
from multiprocessing import cpu_count
from medcat.utils.hasher import Hasher
from typing import Optional, Iterable, Tuple, Dict, Any


def weighted_average(step: int, factor: float) -> float:
    return max(0.1, 1 - (step ** 2 * factor))


def workers(workers_override: Optional[int] = None) -> int:
    return max(cpu_count() - 1, 1) if workers_override is None else workers_override


class ConfigMixin(object):

    jsonpickle.set_encoder_options('json', sort_keys=True, indent=2)

    def __iter__(self) -> Iterable[Tuple]:
        for attr, value in self.__dict__.items():
            yield attr, value

    def save(self, save_path: str) -> None:
        r''' Save the config into a .json file

        Args:
            save_path (`str`):
                Where to save the created json file
        '''
        # We want to save the dict here, not the whole class
        json_string = jsonpickle.encode(self.__dict__)

        with open(save_path, 'w') as f:
            f.write(json_string)

    def merge_config(self, config_dict: Dict) -> None:
        r''' Merge a config_dict with the existing config object.

        Args:
            config_dict (`dict`):
                A dictionary which key/values should be added to this class.
        '''
        for key in config_dict.keys():
            if key in self.__dict__ and isinstance(self.__dict__[key], dict):
                self.__dict__[key].update(config_dict[key])
            else:
                self.__dict__[key] = config_dict[key]

    def parse_config_file(self, path: str) -> None:
        r'''
        Parses a configuration file in text format. Must be like:
                cat.<variable>.<key> = <value>
                ...
            Where:
                variable: linking, general, ner, ...
                key: a key in the config dict e.g. subsample_after for linking
                value: the value for the key, will be parsed with `eval`
        '''
        with open(path, 'r') as f:
            for line in f:
                if line.strip() and line.startswith("cat."):
                    line = line[4:]
                    left, right = line.split("=")
                    variable, key = left.split(".")
                    variable = variable.strip()
                    key = key.strip()
                    value = eval(right.strip())

                    attr = getattr(self, variable)
                    attr[key] = value

        self.rebuild_re()

    def rebuild_re(self) -> None:
        pass

    def get_hash(self):
        hasher = Hasher()
        for k, v in self.__dict__.items():
            hasher.update(v)

        return hasher.hexdigest()

    def __str__(self) -> str:
        json_obj = {}
        for attr, value in self:    # type: ignore
            json_obj[attr] = value
        return jsonpickle.encode(json_obj)

    @classmethod
    def load(cls, save_path: str) -> "ConfigMixin":
        r''' Load config from a json file, note that fields that
        did not exist in the old config but do exist in the current
        version of the ConfigMetaCAT class will be kept.

        Args:
            save_path (`str`):
                Path to the json file to load
        '''
        config = cls()

        # Read the jsonpickle string
        with open(save_path) as f:
            config_dict = jsonpickle.decode(f.read())

        config.merge_config(config_dict)

        return config

    @classmethod
    def from_dict(cls, config_dict: Dict) -> "ConfigMixin":
        config = cls()
        config.merge_config(config_dict)

        return config


class Config(ConfigMixin):

    def __init__(self) -> None:
        self.version: Dict[str, Any] = {
            'id': None, # Will be: hash of most things 
            'last_modified': None, # Yep
            'location': None, # Path/URL/Whatever to where is this CDB located
            'history': [], # Populated automatically
            'description': "No description", # General description and what it was trained on
            'meta_cats': {}, # Populated automatically
            'cdb_info': {}, # Populated automatically, output from cdb.print_stats
            'performance': {'ner': {}, 'meta': {}}, # NER general performance, meta should be: {'meta': {'model_name': {'f1': <>, 'p': <>, ...}, ...}}
            'ontology': None, # What was used to build the CDB, e.g. SNOMED_202009
            'medcat_version': None, # Which version of medcat was used to build the CDB
        }

        # CDB Maker
        self.cdb_maker: Dict[str, Any] = {
                # If multiple names or type_ids for a concept present in one row of a CSV, they are separted
                # by the character below.
                'multi_separator': '|',
                # Name versions to be generated.
                'name_versions': ['LOWER', 'CLEAN'],
                # Should preferred names with parenthesis be cleaned 0 means no, else it means if longer than or equal
                # e.g. Head (Body part) -> Head
                'remove_parenthesis': 5,
                # Minimum number of letters required in a name to be accepted for a concept
                'min_letters_required': 2,
                }

        # Used mainly to configure the output of the get_entities function, and in that also the output of
        #get_json and multiprocessing
        self.annotation_output: Dict[str, Any] = {
                'doc_extended_info': False,
                'context_left': -1,
                'context_right': -1,
                'lowercase_context': True,
                'include_text_in_output': False,
                }

        self.general: Dict[str, Any] = {
                # Logging config for everything | 'tagger' can be disabled, but will cause a drop in performance
                'log_level': logging.INFO,
                'log_format': '%(levelname)s:%(name)s: %(message)s',
                'log_path': './medcat.log',
                'spacy_disabled_components': ['ner', 'parser', 'vectors', 'textcat',
                                              'entity_linker', 'sentencizer', 'entity_ruler', 'merge_noun_chunks',
                                              'merge_entities', 'merge_subtokens'],
                # What model will be used for tokenization
                'spacy_model': 'en_core_web_md',
                # Separator that will be used to merge tokens of a name. Once a CDB is built this should
                #always stay the same.
                'separator': '~',
                # Should we check spelling - note that this makes things much slower, use only if necessary. The only thing necessary
                #for the spell checker to work is vocab.dat and cdb.dat built with concepts in the respective language.
                'spell_check': True,
                # Should we process diacritics - for languages other than English, symbols such as 'é, ë, ö' can be relevant.
                # Note that this makes spell_check slower.
                'diacritics': False,
                # If True the spell checker will try harder to find mistakes, this can slow down
                #things drastically.
                'spell_check_deep': False,
                # Spelling will not be checked for words with length less than this
                'spell_check_len_limit': 7,
                # If set to True functions like get_entities and get_json will return nested_entities and overlaps
                'show_nested_entities': False,
                # When unlinking a name from a concept should we do full_unlink (means unlink a name from all concepts, not just the one in question)
                'full_unlink': False,
                # Number of workers used by a parallelizable pipeline component
                'workers': workers(),
                # Should the labels of entities (shown in displacy) be pretty or just 'concept'. Slows down the annotation pipeline
                #should not be used when annotating millions of documents. If `None` it will be the string "concept", if `short` it will be CUI,
                #if `long` it will be CUI | Name | Confidence
                'make_pretty_labels': None,
                # If the cdb.addl_info['cui2group'] is provided and this option enabled, each CUI will be maped to the group
                'map_cui_to_group': False,
                # Checkpointing config
                'checkpoint': {
                    # When doing training this is the name of the directory where checkpoints will be saved
                    'output_dir': 'checkpoints',
                    # When training how often to save the checkpoint (one step represents one document), if None no ckpts will be created
                    'steps': None,
                    # When training the maximum checkpoints will be kept on the disk
                    "max_to_keep": 1,
                },
                }

        self.preprocessing: Dict[str, Any] = {
                # Should stopwords be skipped/ingored when processing input
                'skip_stopwords': False,
                # This words will be completly ignored from concepts and from the text (must be a Set)
                'words_to_skip': set(['nos']),
                # All punct will be skipped by default, here you can set what will be kept
                'keep_punct': {'.', ':'},
                # Nothing below this length will ever be normalized (input tokens or concept names), normalized means lemmatized in this case
                'min_len_normalize': 5,
                # If None the default set of stowords from spacy will be used. This must be a Set.
                'stopwords': None,
                # Documents longer  than this will be trimmed
                'max_document_length': 1000000,
                # Should specific word types be normalized: e.g. running -> run
                'do_not_normalize': {'VBD', 'VBG', 'VBN', 'VBP', 'JJS', 'JJR'},
                }

        self.ner: Dict[str, Any] = {
                # Do not detect names below this limit, skip them
                'min_name_len': 3,
                # When checkng tokens for concepts you can have skipped tokens inbetween
                #used ones (usually spaces, new lines etc). This number tells you how many skipped can you have.
                'max_skip_tokens': 2,
                # Check uppercase to distinguish uppercase and lowercase words that have a different meaning.
                'check_upper_case_names': False,
                # Any name shorter than this must be uppercase in the text to be considered. If it is not uppercase
                #it will be skipped.
                'upper_case_limit_len': 4,
                # Try reverse word order for short concepts (2 words max), e.g. heart disease -> disease heart
                'try_reverse_word_order': False,
                }

        self.linking: Dict[str, Any] = {
                # Should it train or not, this is set automatically ignore in 99% of cases and do not set manually
                'train': True,
                # Linear anneal
                'optim': {'type': 'linear', 'base_lr': 1, 'min_lr': 0.00005},
                # If <1 during unsupervised training the detected term will be randomly replaced with a probability of 1 - random_replacement_unsupervised
                #Replaced with a synonym used for that term
                'random_replacement_unsupervised': 0.80,
                # 'optim': {'type': 'standard', 'lr': 1},
                # 'optim': {'type': 'moving_avg', 'alpha': 0.99, 'e': 1e-4, 'size': 100},
                # All concepts below this will always be disambiguated
                'disamb_length_limit': 3,
                # Context vector sizes that will be calculated and used for linking
                'context_vector_sizes': {'xlong': 27, 'long': 18, 'medium': 9, 'short': 3},
                # Weight of each vector in the similarity score - make trainable at some point. Should add up to 1.
                'context_vector_weights': {'xlong': 0.1, 'long': 0.4, 'medium': 0.4, 'short': 0.1},
                # If True it will filter before doing disamb. Useful for the trainer.
                'filter_before_disamb': False,
                # Concepts that have seen less training examples than this will not be used for
                #similarity calculation and will have a similarity of -1.
                'train_count_threshold': 1,
                # Do we want to calculate context similarity even for concepts that are not ambigous.
                'always_calculate_similarity': False,
                # Weights for a weighted average
                #'weighted_average_function': partial(weighted_average, factor=0.02),
                'weighted_average_function': partial(weighted_average, factor=0.0004),
                # Concepts below this similarity will be ignored. Type can be static/dynamic - if dynamic each CUI has a different TH
                #and it is calcualted as the average confidence for that CUI * similarity_threshold. Take care that dynamic works only
                #if the cdb was trained with calculate_dynamic_threshold = True.
                'calculate_dynamic_threshold': False,
                'similarity_threshold_type': 'static',
                'similarity_threshold': 0.25,
                # Probability for the negative context to be added for each positive addition
                'negative_probability': 0.5,
                # Do we ignore punct/num when negative sampling
                'negative_ignore_punct_and_num': True,
                # If >0 concepts for which a detection is its primary name will be preferred by that amount (0 to 1)
                'prefer_primary_name': 0.35,
                # If >0 concepts that are more frequent will be prefered by a multiply of this amount
                'prefer_frequent_concepts': 0.35,
                # DISABLED in code permanetly: Subsample during unsupervised training if a concept has received more than
                'subsample_after': 30000,
                # When adding a positive example, should it also be treated as Negative for concepts
                #which link to the postive one via names (ambigous names).
                'devalue_linked_concepts': False,
                # If true when the context of a concept is calculated (embedding) the words making that concept are not taken into accout
                'context_ignore_center_tokens': False,
                # Filters
                'filters': {
                    'cuis': set(), # CUIs in this filter will be included, everything else excluded, must be a set, if empty all cuis will be included
                    },
                }


        # Some regex that we will need
        self.word_skipper = re.compile('^({})$'.format('|'.join(self.preprocessing['words_to_skip'])))
        # Very agressive punct checker, input will be lowercased
        self.punct_checker = re.compile(r'[^a-z0-9]+')

    # Override
    def rebuild_re(self) -> None:
        # Some regex that we will need
        self.word_skipper = re.compile('^({})$'.format('|'.join(self.preprocessing['words_to_skip'])))
        # Very agressive punct checker, input will be lowercased
        self.punct_checker = re.compile(r'[^a-z0-9]+')

    def get_hash(self):
        hasher = Hasher()
        for k, v in self.__dict__.items():
            if k not in ['version', 'general', 'linking']:
                hasher.update(v, length=True)
            elif k == 'general':
                for k2, v2 in v.items():
                    if k2 != 'spacy_model':
                        hasher.update(v2, length=False)
                    else:
                        # Ignore spacy model
                        pass
            elif k == 'linking':
                for k2, v2 in v.items():
                    if k2 != "filters":
                        hasher.update(v2, length=False)
                    else:
                        hasher.update(v2, length=True)

        return hasher.hexdigest()



import os
import json
import logging
import torch
import numpy
from multiprocessing import Lock
from torch import nn, Tensor
from spacy.tokens import Doc
from datetime import datetime
from typing import Iterable, Iterator, Optional, Dict, List, Tuple, cast, Union
from medcat.utils.hasher import Hasher
from medcat.config_meta_cat import ConfigMetaCAT
from medcat.utils.meta_cat.ml_utils import predict, train_model, set_all_seeds, eval_model
from medcat.utils.meta_cat.data_utils import prepare_from_json, encode_category_values
from medcat.utils.loggers import add_handlers
from medcat.pipeline.pipe_runner import PipeRunner
from medcat.tokenizers.meta_cat_tokenizers import TokenizerWrapperBase
from medcat.utils.meta_cat.data_utils import Doc as FakeDoc

# It should be safe to do this always, as all other multiprocessing
#will be finished before data comes to meta_cat
os.environ["TOKENIZERS_PARALLELISM"] = "true"


class MetaCAT(PipeRunner):
    r''' TODO: Add documentation
    '''

    # Custom pipeline component name
    name = 'meta_cat'
    _component_lock = Lock()

    log = logging.getLogger(__package__)
    # Add file and console handlers
    log = add_handlers(log)

    # Override
    def __init__(self,
                 tokenizer: Optional[TokenizerWrapperBase] = None,
                 embeddings: Optional[Union[Tensor, numpy.ndarray]] = None,
                 config: Optional[ConfigMetaCAT] = None) -> None:
        if config is None:
            config = ConfigMetaCAT()
        self.config = config
        set_all_seeds(config.general['seed'])

        if tokenizer is not None:
            # Set it in the config
            config.general['tokenizer_name'] = tokenizer.name
            config.general['vocab_size'] = tokenizer.get_size()
            # We will also set the padding
            config.model['padding_idx'] = tokenizer.get_pad_id()
        self.tokenizer = tokenizer

        self.embeddings = torch.tensor(embeddings, dtype=torch.float32) if embeddings is not None else None
        self.model = self.get_model(embeddings=self.embeddings)

    def get_model(self, embeddings: Optional[Tensor]) -> nn.Module:
        config = self.config
        if config.model['model_name'] == 'lstm':
            from medcat.utils.meta_cat.models import LSTM
            model = LSTM(embeddings, config)
        else:
            raise ValueError("Unknown model name %s" % config.model['model_name'])

        return model

    def get_hash(self):
        r''' A partial hash trying to catch differences between models
        '''
        hasher = Hasher()
        # Set last_train_on if None
        if self.config.train['last_train_on'] is None:
            self.config.train['last_train_on'] = datetime.now().timestamp()

        hasher.update(self.config.get_hash())
        return hasher.hexdigest()

    def train(self, json_path: str, save_dir_path: Optional[str] = None) -> Dict:
        r''' Train or continue training a model give a json_path containing a MedCATtrainer export. It will
        continue training if an existing model is loaded or start new training if the model is blank/new.

        Args:
            json_path (`str`):
                Path to a MedCATtrainer export containing the meta_annotations we want to train for.
            save_dir_path (`str`, optional, defaults to `None`):
                In case we have aut_save_model (meaning during the training the best model will be saved)
                we need to set a save path.
        '''
        g_config = self.config.general
        t_config = self.config.train

        # Load the medcattrainer export
        with open(json_path, 'r') as f:
            data_loaded: Dict = json.load(f)

        # Create directories if they don't exist
        if t_config['auto_save_model']:
            if save_dir_path is None:
                raise Exception("The `save_dir_path` argument is required if `aut_save_model` is `True` in the config")
            else:
                os.makedirs(save_dir_path, exist_ok=True)

        # Prepare the data
        assert self.tokenizer is not None
        data = prepare_from_json(data_loaded, g_config['cntx_left'], g_config['cntx_right'], self.tokenizer, cui_filter=t_config['cui_filter'],
                replace_center=g_config['replace_center'], prerequisites=t_config['prerequisites'],
                lowercase=g_config['lowercase'])

        # Check is the name there
        category_name = g_config['category_name']
        if category_name not in data:
            raise Exception("The category name does not exist in this json file. You've provided '{}', while the possible options are: {}".format(
                category_name, " | ".join(list(data.keys()))))

        data = data[category_name]

        category_value2id = g_config['category_value2id']
        if not category_value2id:
            # Encode the category values
            data, category_value2id = encode_category_values(data)
            g_config['category_value2id'] = category_value2id
        else:
            # We already have everything, just get the data
            data, _ = encode_category_values(data, existing_category_value2id=category_value2id)

        # Make sure the config number of classes is the same as the one found in the data
        if len(category_value2id) != self.config.model['nclasses']:
            self.log.warning("The number of classes set in the config is not the same as the one found in the data: {} vs {}".format(
                             self.config.model['nclasses'], len(category_value2id)))
            self.log.warning("Auto-setting the nclasses value in config and rebuilding the model.")
            self.config.model['nclasses'] = len(category_value2id)
            self.model = self.get_model(embeddings=self.embeddings)

        report = train_model(self.model, data=data, config=self.config, save_dir_path=save_dir_path)

        # If autosave, then load the best model here
        if t_config['auto_save_model']:
            if save_dir_path is None:
                raise Exception("The `save_dir_path` argument is required if `aut_save_model` is `True` in the config")
            else:
                path = os.path.join(save_dir_path, 'model.dat')
                device = torch.device(g_config['device'])
                self.model.load_state_dict(torch.load(path, map_location=device))

                # Save everything now
                self.save(save_dir_path=save_dir_path)

        self.config.train['last_train_on'] = datetime.now().timestamp()
        return report

    def eval(self, json_path: str) -> Dict:
        g_config = self.config.general
        t_config = self.config.train

        with open(json_path, 'r') as f:
            data_loaded: Dict = json.load(f)

        # Prepare the data
        assert self.tokenizer is not None
        data = prepare_from_json(data_loaded, g_config['cntx_left'], g_config['cntx_right'], self.tokenizer, cui_filter=t_config['cui_filter'],
                replace_center=g_config['replace_center'], prerequisites=t_config['prerequisites'],
                lowercase=g_config['lowercase'])

        # Check is the name there
        category_name = g_config['category_name']
        if category_name not in data:
            raise Exception("The category name does not exist in this json file.")

        data = data[category_name]

        # We already have everything, just get the data
        category_value2id = g_config['category_value2id']
        data, _ = encode_category_values(data, existing_category_value2id=category_value2id)

        # Run evaluation
        assert self.tokenizer is not None
        result = eval_model(self.model, data, config=self.config, tokenizer=self.tokenizer)

        return result

    def save(self, save_dir_path: str) -> None:
        r''' Save all components of this class to a file

        Args:
            save_dir_path(`str`):
                Path to the directory where everything will be saved.
        '''
        # Create dirs if they do not exist
        os.makedirs(save_dir_path, exist_ok=True)

        # Save tokenizer
        assert self.tokenizer is not None
        self.tokenizer.save(save_dir_path)

        # Save config
        self.config.save(os.path.join(save_dir_path, 'config.json'))

        # Save the model
        model_save_path = os.path.join(save_dir_path, 'model.dat')
        torch.save(self.model.state_dict(), model_save_path)

        # This is everything we need to save from the class, we do not
        #save the class itself.

    @classmethod
    def load(cls, save_dir_path: str, config_dict: Optional[Dict] = None) -> "MetaCAT":
        r''' Load a meta_cat object.

        Args:
            save_dir_path (`str`):
                The directory where all was saved.
            config_dict (`dict`):
                This can be used to overwrite saved parameters for this meta_cat
                instance. Why? It is needed in certain cases where we autodeploy stuff.

        Returns:
            meta_cat (`medcat.MetaCAT`):
                You don't say
        '''

        # Load config
        config = cast(ConfigMetaCAT, ConfigMetaCAT.load(os.path.join(save_dir_path, 'config.json')))

        # Overwrite loaded paramters with something new
        if config_dict is not None:
            config.merge_config(config_dict)

        tokenizer: Optional[TokenizerWrapperBase] = None
        # Load tokenizer (TODO: This should be converted into a factory or something)
        if config.general['tokenizer_name'] == 'bbpe':
            from medcat.tokenizers.meta_cat_tokenizers import TokenizerWrapperBPE
            tokenizer = TokenizerWrapperBPE.load(save_dir_path)
        elif config.general['tokenizer_name'] == 'bert-tokenizer':
            from medcat.tokenizers.meta_cat_tokenizers import TokenizerWrapperBERT
            tokenizer = TokenizerWrapperBERT.load(save_dir_path)

        # Create meta_cat
        meta_cat = cls(tokenizer=tokenizer, embeddings=None, config=config)

        # Load the model
        model_save_path = os.path.join(save_dir_path, 'model.dat')
        device = torch.device(config.general['device'])
        if not torch.cuda.is_available() and device.type == 'cuda':
            MetaCAT.log.warning('Loading a MetaCAT model without GPU availability, stored config used GPU')
            config.general['device'] = 'cpu'
            device = torch.device('cpu')
        meta_cat.model.load_state_dict(torch.load(model_save_path, map_location=device))

        return meta_cat

    def prepare_document(self, doc: Doc, input_ids: List, offset_mapping: List, lowercase: bool) -> Tuple:
        r'''

        Args:
            doc - spacy
            input_ids
            offset_mapping
        '''
        config = self.config
        cntx_left = config.general['cntx_left']
        cntx_right = config.general['cntx_right']
        replace_center = config.general['replace_center']

        # Should we annotate overlapping entities
        if config.general['annotate_overlapping']:
            ents = doc._.ents
        else:
            ents = doc.ents

        samples = []
        last_ind = 0
        ent_id2ind = {} # Map form entitiy ID to where is it in the samples array
        for ent in sorted(ents, key=lambda ent: ent.start_char):
            start = ent.start_char
            end = ent.end_char

            ind = 0
            # Start where the last ent was found, cannot be before it as we've sorted
            for ind, pair in enumerate(offset_mapping[last_ind:]):
                if start >= pair[0] and start < pair[1]:
                    break
            ind = last_ind + ind # If we did not start from 0 in the for loop
            last_ind = ind

            _start = max(0, ind - cntx_left)
            _end = min(len(input_ids), ind + 1 + cntx_right)
            tkns = input_ids[_start:_end]
            cpos = cntx_left + min(0, ind-cntx_left)

            if replace_center is not None:
                if lowercase:
                    replace_center = replace_center.lower()
                # We start from ind
                s_ind = ind
                e_ind = ind
                for _ind, pair in enumerate(offset_mapping[ind:]):
                    if end > pair[0] and end <= pair[1]:
                        e_ind = _ind + ind
                        break
                ln = e_ind - s_ind # Length of the concept in tokens
                assert self.tokenizer is not None
                tkns = tkns[:cpos] + self.tokenizer(replace_center)['input_ids'] + tkns[cpos+ln+1:]

            samples.append([tkns, cpos])
            ent_id2ind[ent._.id] = len(samples) - 1

        return ent_id2ind, samples

    @staticmethod
    def batch_generator(stream: Iterable[Doc], batch_size_chars: int) -> Iterable[List[Doc]]:
        docs = []
        char_count = 0
        for doc in stream:
            char_count += len(doc.text)
            docs.append(doc)
            if char_count < batch_size_chars:
                continue
            yield docs
            docs = []
            char_count = 0

        # If there is anything left return that also
        if len(docs) > 0:
            yield docs

    # Override
    def pipe(self, stream: Iterable[Union[Doc, FakeDoc]], *args, **kwargs) -> Iterator[Doc]:
        r''' Process many documents at once.

        Args:
            stream (Iterable[spacy.tokens.Doc]):
                List of spacy documents.
        '''
        # Just in case
        if stream is None or not stream:
            return stream

        config = self.config
        id2category_value = {v: k for k, v in config.general['category_value2id'].items()}
        batch_size_chars = config.general['pipe_batch_size_in_chars']

        if config.general['device'] == 'cpu' or config.general['disable_component_lock']:
            yield from self._set_meta_anns(stream, batch_size_chars, config, id2category_value)
        else:
            with MetaCAT._component_lock:
                yield from self._set_meta_anns(stream, batch_size_chars, config, id2category_value)

    def _set_meta_anns(self,
                       stream: Iterable[Union[Doc, FakeDoc]],
                       batch_size_chars: int,
                       config: ConfigMetaCAT,
                       id2category_value: Dict) -> Iterator[Optional[Doc]]:
        for docs in self.batch_generator(stream, batch_size_chars):
            try:
                if not config.general['save_and_reuse_tokens'] or docs[0]._.share_tokens is None:
                    if config.general['lowercase']:
                        all_text = [doc.text.lower() for doc in docs]
                    else:
                        all_text = [doc.text for doc in docs]
                    assert self.tokenizer is not None
                    all_text_processed = self.tokenizer(all_text)
                    doc_ind2positions = {}
                    data: List = [] # The thing that goes into the model
                    for i, doc in enumerate(docs):
                        ent_id2ind, samples = self.prepare_document(doc, input_ids=all_text_processed[i]['input_ids'],
                                                                    offset_mapping=all_text_processed[i]['offset_mapping'],
                                                                    lowercase=config.general['lowercase'])
                        doc_ind2positions[i] = (len(data), len(data) + len(samples), ent_id2ind) # Needed so we know where is what in the big data array
                        data.extend(samples)
                        if config.general['save_and_reuse_tokens']:
                            doc._.share_tokens = (samples, doc_ind2positions[i])
                else:
                    # This means another model has already processed the data and we can just use it. This is a
                    #dangerous option - as it assumes the other model has the same tokenizer and context size.
                    data = []
                    doc_ind2positions = {}
                    for i, doc in enumerate(docs):
                        data.extend(doc._.share_tokens[0])
                        doc_ind2positions[i] = doc._.share_tokens[1]

                all_predictions, all_confidences = predict(self.model, data, config)
                for i, doc in enumerate(docs):
                    start_ind, end_ind, ent_id2ind = doc_ind2positions[i]

                    predictions = all_predictions[start_ind:end_ind]
                    confidences = all_confidences[start_ind:end_ind]
                    if config.general['annotate_overlapping']:
                        ents = doc._.ents
                    else:
                        ents = doc.ents

                    for ent in ents:
                        ent_ind = ent_id2ind[ent._.id]
                        value = id2category_value[predictions[ent_ind]]
                        confidence = confidences[ent_ind]
                        if ent._.meta_anns is None:
                            ent._.meta_anns = {config.general['category_name']: {'value': value,
                                                                                 'confidence': float(confidence),
                                                                                 'name': config.general['category_name']}}
                        else:
                            ent._.meta_anns[config.general['category_name']] = {'value': value,
                                                                                'confidence': float(confidence),
                                                                                'name': config.general['category_name']}
                yield from docs
            except Exception as e:
                self.get_error_handler()(self.name, self, docs, e)
                yield from [None] * len(docs)

    # Override
    def __call__(self, doc: Doc) -> Doc:
        ''' Process one document, used in the spacy pipeline for sequential
        document processing.

        Args:
            doc (spacy.tokens.Doc):
                A spacy document
        '''

        # Just call the pipe method
        doc = next(self.pipe(iter([doc])))

        return doc



import unittest
from medcat.cat import CAT
from medcat.config import Config
from medcat.utils.model_creator import create_models
from pathlib import Path


class EntityLinkingTest(unittest.TestCase):
    """Test entity linking after creating MedCAT CDB and Vocab models.

    During class setup, MedCAT CDB and Vocabulary models are generated from included test data. Subsequently, these
    models are assessed using various entity linking scenarios.
    """

    @classmethod
    def setUpClass(cls):
        cdb, vocab = create_models(Path('tests/model_creator/config_example.yml'))
        medcat_config = Config()
        medcat_config.parse_config_file(Path('tests/model_creator/medcat.txt'))
        cls.cat = CAT(cdb=cdb, vocab=vocab, config=medcat_config)
        cls.cdb = cdb

    def assert_linked_entities(self, doc, expected_entities, unexpected_entities=None):
        """General assertion function to assess linked entities.

        Function to assess whether a doc-object's linked entities are in concordance with a list of expected entities.

        Args:
            doc (spacy.tokens.Doc):
                A spaCy doc-object containing entities and their respective CUIs as provided by MedCAT
            expected_entities (list):
                List of CUIs that should be in the doc object.
            unexpected_entities (list):
                List of CUIs that should not be in the doc object.
        """
        # Extract found entities
        linked_entities = [entity._.cui for entity in doc.ents]

        # Assert whether expected entities are found
        for expected_entity in expected_entities:
            self.assertIn(expected_entity, linked_entities)

        # Assert whether unexpected entities are not found
        if unexpected_entities is not None:
            for unexpected_entity in unexpected_entities:
                self.assertNotIn(unexpected_entity, linked_entities)


class TestEntityLinking(EntityLinkingTest):
    """Test entity linking.

    Test the created medcat models, as well as the entity linking method, by assessing whether a number of expected
    entities are found in a text.
    """

    def test_entity_linking(self):
        text = "Common treatments include surgery, chemotherapy, and radiotherapy. NSCLC is sometimes treated with " \
               "surgery, whereas SCLC usually responds better to chemotherapy and radiotherapy. Of all people with " \
               "lung cancer in the US, around 17% to 20% survive for at least five years after diagnosis."

        # These entities should be found in above text
        expected_entities = ['C1522449',  # surgery
                             'C3665472',  # chemotherapy
                             'C1522449',  # radiotherapy
                             'C0007131',  # NSCLC
                             'C0149925',  # SCLC
                             'C0242379']  # lung cancer
        unexpected_entities = ['C0006826']  # cancer
        doc = self.cat(text)
        self.assert_linked_entities(doc, expected_entities, unexpected_entities)


class TestLemmatization(EntityLinkingTest):
    """Test lemmatization.

    Some entities should be lemmatized before they can be linked to an entity. The minimal number of
    characters for a word required to be lemmatized is defined in the "min_len_normalize" configuration.
    """

    def test_lemmatization_not_required(self):
        expected_entities = ['C0085639']  # Fall
        text = "In the event of a fall"
        doc = self.cat(text)
        self.assert_linked_entities(doc, expected_entities)

    def test_lemmatization_required(self):
        expected_entities = ['C0085639']  # Fall
        text = "The patient is afraid of falling."
        doc = self.cat(text)
        self.assert_linked_entities(doc, expected_entities)


class TestDiacritics(EntityLinkingTest):
    """Test handling of diacritics.

    Diacritics are common in many non-English languages and require specific functionality in MedCAT. This
    functionality can be enabled via the "diacritics" configuration.
    """

    def test_diacritics_in_cdb(self):
        self.assertIn('ménière', self.cdb.cui2snames['C0025281'])

    def test_diacritics_in_text(self):
        expected_entities = ['C0025281']
        text = "Ménière's disease (MD) is a disorder of the inner ear that is characterized by episodes of vertigo, " \
               "tinnitus, hearing loss, and a fullness in the ear."
        doc = self.cat(text)
        self.assert_linked_entities(doc, expected_entities)


class TestCheckUpperCaseNames(EntityLinkingTest):
    """ Test the "check_upper_case_names" functionality.

    Some capitalized words, often acronyms, have a different meaning when written in lowercase.
    Distinguishing these words can be done via by using the check_upper_case_names functionality, which can be
    enabled via the "check_upper_case_names" configuration.
    """

    def test_uppercase_abbreviation_in_sentence(self):
        text = "Phosphorylation of the MAP has an effect."
        expected_entities = ['C0026045']  # MAP (Microtubule-Associated Proteins)
        doc = self.cat(text)
        self.assert_linked_entities(doc, expected_entities)

    def test_uppercase_abbreviation_start_sentence(self):
        text = "MAP groups have been identified."
        expected_entities = ['C0026045']  # MAP (Microtubule-Associated Proteins)
        doc = self.cat(text)
        self.assert_linked_entities(doc, expected_entities)

    def test_lowercase_non_abbreviation_in_sentence(self):
        text = "MedCAT should not map this name to a medical concept."
        doc = self.cat(text)
        self.assertEqual(len(doc.ents), 0)

    def test_lowercase_non_abbreviation_start_sentence(self):
        text = "Map should not be linked to a medical concept."
        doc = self.cat(text)
        self.assertEqual(len(doc.ents), 0)

    def test_uppercase_non_abbreviation(self):
        text = "TAXOTERE is a drug used to treat certain types of cancer."
        expected_entities = ['C0699967']
        doc = self.cat(text)
        self.assert_linked_entities(doc, expected_entities)


if __name__ == '__main__':
    unittest.main()



import unittest
import logging
import os
import requests
from spacy.language import Language
from medcat.cdb import CDB
from medcat.vocab import Vocab
from medcat.config import Config
from medcat.pipe import Pipe
from medcat.meta_cat import MetaCAT
from medcat.preprocessing.taggers import tag_skip_and_punct
from medcat.preprocessing.tokenizers import spacy_split_all
from medcat.utils.normalizers import BasicSpellChecker, TokenNormalizer
from medcat.ner.vocab_based_ner import NER
from medcat.linking.context_based_linker import Linker
from medcat.tokenizers.meta_cat_tokenizers import TokenizerWrapperBERT
from transformers import AutoTokenizer



class PipeTests(unittest.TestCase):

    @classmethod
    def setUpClass(cls) -> None:
        cls.config = Config()
        cls.config.general['log_level'] = logging.INFO
        cls.config.general["spacy_model"] = "en_core_web_md"
        cls.config.ner['max_skip_tokens'] = 1
        cls.config.ner['upper_case_limit_len'] = 4
        cls.config.linking['disamb_length_limit'] = 2
        cls.cdb = CDB(config=cls.config)

        vocab_path = "./tmp_vocab.dat"
        if not os.path.exists(vocab_path):
            tmp = requests.get("https://medcat.rosalind.kcl.ac.uk/media/vocab.dat")
            with open(vocab_path, 'wb') as f:
                f.write(tmp.content)

        cls.vocab = Vocab.load(vocab_path)
        cls.spell_checker = BasicSpellChecker(cdb_vocab=cls.cdb.vocab, config=cls.config, data_vocab=cls.vocab)
        cls.ner = NER(cls.cdb, cls.config)
        cls.linker = Linker(cls.cdb, cls.vocab, cls.config)

        _tokenizer = TokenizerWrapperBERT(hf_tokenizers=AutoTokenizer.from_pretrained("bert-base-uncased"))
        cls.meta_cat = MetaCAT(tokenizer=_tokenizer)

        cls.text = "CDB - I was running and then Movar Virus attacked and CDb"
        cls.undertest = Pipe(tokenizer=spacy_split_all, config=cls.config)

    @classmethod
    def tearDownClass(cls) -> None:
        cls.undertest.destroy()

    def setUp(self) -> None:
        PipeTests.undertest.force_remove(tag_skip_and_punct.name)
        PipeTests.undertest.force_remove(TokenNormalizer.name)
        PipeTests.undertest.force_remove(PipeTests.ner.name)
        PipeTests.undertest.force_remove(PipeTests.linker.name)
        PipeTests.undertest.force_remove(PipeTests.meta_cat.name)

    def test_add_tagger(self):
        PipeTests.undertest.add_tagger(tagger=tag_skip_and_punct, name=tag_skip_and_punct.name, additional_fields=["is_punct"])

        self.assertEqual(tag_skip_and_punct.name, Language.get_factory_meta(tag_skip_and_punct.name).factory)
        self.assertEqual(PipeTests.config, Language.get_factory_meta(tag_skip_and_punct.name).default_config["config"])

    def test_add_token_normalizer(self):
        PipeTests.undertest.add_token_normalizer(PipeTests.config, spell_checker=PipeTests.spell_checker)

        self.assertEqual(TokenNormalizer.name, Language.get_factory_meta(TokenNormalizer.name).factory)

    def test_add_ner(self):
        PipeTests.undertest.add_ner(PipeTests.ner)

        self.assertEqual(PipeTests.ner.name, Language.get_factory_meta(PipeTests.ner.name).factory)

    def test_add_linker(self):
        PipeTests.undertest.add_linker(PipeTests.linker)

        self.assertEqual(PipeTests.linker.name, Language.get_factory_meta(PipeTests.linker.name).factory)

    def test_add_meta_cat(self):
        PipeTests.undertest.add_meta_cat(PipeTests.meta_cat)

        self.assertEqual(PipeTests.meta_cat.name, Language.get_factory_meta(PipeTests.meta_cat.name).factory)

    def test_batch_multi_process(self):
        PipeTests.undertest.add_tagger(tagger=tag_skip_and_punct, additional_fields=["is_punct"])
        PipeTests.undertest.add_token_normalizer(PipeTests.config, spell_checker=PipeTests.spell_checker)
        PipeTests.undertest.add_ner(PipeTests.ner)
        PipeTests.undertest.add_linker(PipeTests.linker)
        PipeTests.undertest.add_meta_cat(PipeTests.meta_cat)

        PipeTests.undertest.set_error_handler(_error_handler)
        docs = list(self.undertest.batch_multi_process([PipeTests.text, PipeTests.text, PipeTests.text], n_process=1, batch_size=1))
        PipeTests.undertest.reset_error_handler()

        self.assertEqual(3, len(docs))
        self.assertEqual(PipeTests.text, docs[0].text)
        self.assertEqual(PipeTests.text, docs[1].text)
        self.assertEqual(PipeTests.text, docs[2].text)

    def test_callable_with_generated_texts(self):
        def _generate_texts(texts):
            yield from texts

        PipeTests.undertest.add_tagger(tagger=tag_skip_and_punct, additional_fields=["is_punct"])
        PipeTests.undertest.add_token_normalizer(PipeTests.config, spell_checker=PipeTests.spell_checker)
        PipeTests.undertest.add_ner(PipeTests.ner)
        PipeTests.undertest.add_linker(PipeTests.linker)
        PipeTests.undertest.add_meta_cat(PipeTests.meta_cat)

        docs = list(self.undertest(_generate_texts([PipeTests.text, None, PipeTests.text])))

        self.assertEqual(3, len(docs))
        self.assertEqual(PipeTests.text, docs[0].text)
        self.assertIsNone(docs[1])
        self.assertEqual(PipeTests.text, docs[2].text)

    def test_callable_with_single_text(self):
        PipeTests.undertest.add_tagger(tagger=tag_skip_and_punct, additional_fields=["is_punct"])
        PipeTests.undertest.add_token_normalizer(PipeTests.config, spell_checker=PipeTests.spell_checker)
        PipeTests.undertest.add_ner(PipeTests.ner)
        PipeTests.undertest.add_linker(PipeTests.linker)
        PipeTests.undertest.add_meta_cat(PipeTests.meta_cat)

        doc = self.undertest(PipeTests.text)

        self.assertEqual(PipeTests.text, doc.text)

    def test_callable_with_multi_texts(self):
        PipeTests.undertest.add_tagger(tagger=tag_skip_and_punct, additional_fields=["is_punct"])
        PipeTests.undertest.add_token_normalizer(PipeTests.config, spell_checker=PipeTests.spell_checker)
        PipeTests.undertest.add_ner(PipeTests.ner)
        PipeTests.undertest.add_linker(PipeTests.linker)
        PipeTests.undertest.add_meta_cat(PipeTests.meta_cat)

        docs = list(self.undertest([PipeTests.text, None, PipeTests.text]))

        self.assertEqual(3, len(docs))
        self.assertEqual(PipeTests.text, docs[0].text)
        self.assertIsNone(docs[1])
        self.assertEqual(PipeTests.text, docs[2].text)

def _error_handler(proc_name, proc, docs, e):
    print("Exception raised when when applying component {}".format(proc_name))


if __name__ == '__main__':
    unittest.main()
